(in-package :user)

;; A very simple DATE object.

(defconstant +january+ 1)
(defconstant +february+ 2)
(defconstant +march+ 3)
(defconstant +april+ 4)
(defconstant +may+ 5)
(defconstant +june+ 6)
(defconstant +july+ 7)
(defconstant +august+ 8)
(defconstant +september+ 9)
(defconstant +october+ 10)
(defconstant +november+ 11)
(defconstant +december+ 12)

(defclass datetime ()
  ((year :accessor date-year :initarg :year)
   (month :accessor date-month :initarg :month)
   (day :accessor date-day :initarg :day)
   (hour :accessor date-hour :initarg :hour)
   (minutes :accessor date-minutes :initarg :minutes)))

(defun make-date (month day year hour minutes)
  (make-instance 'datetime 
    :year year 
    :month month
    :day day
    :hour hour
    :minutes minutes))

;; (setq d (make-date +august+ 16 1962 8 30))

(defmethod print-date ((d datetime) stream)
  (format stream "~A/~A/~A ~A:~2,'0D"
    (date-month d) (date-day d) (date-year d)
    (date-hour d) (date-minutes d)))

(defmethod print-object ((d datetime) stream)
  (if *print-escape*
      (print-unreadable-object (d stream :identity t :type t)
        (print-date d stream))
    (print-date d stream)))

(defun isleapyear (year)
   "there is a leap year every year divisible by four 
   except for years which are both divisible by 100 
   and not divisible by 400"
  ;; Use declarations because otherwise REM is slow.
  (declare (optimize (speed 3) (safety 0)) (fixnum year))
  (cond ((not (zerop (rem year 4))) nil)
        ((and (not (zerop (rem year 400))) (zerop (rem year 100))) nil)
        (t t)))

(defmethod dayofyear ((d datetime))
  (let ((day (date-day d))
        (month (date-month d))
        (leapyear (isleapyear (date-year d))))
    (case month
      (1 day)
      (2 (+ day 31))
      (3 (if leapyear (+ day 60) (+ day 59)))
      (4 (if leapyear (+ day 91) (+ day 90)))
      (5 (if leapyear (+ day 121) (+ day 120)))
      (6 (if leapyear (+ day 152) (+ day 151)))
      (7 (if leapyear (+ day 182) (+ day 181)))
      (8 (if leapyear (+ day 213) (+ day 212)))
      (9 (if leapyear (+ day 244) (+ day 243)))
      (10 (if leapyear (+ day 274) (+ day 273)))
      (11 (if leapyear (+ day 305) (+ day 304)))
      (12 (if leapyear (+ day 335) (+ day 334))))))

(defmethod minuteofday ((d datetime))
  (let ((minute (date-minutes d))
        (hour (date-hour d)))
    (+ minute (* hour 60))))

(defmethod date-lessp ((d1 datetime) (d2 datetime))
  (cond ((< (date-year d1) (date-year d2)) t)
        ((> (date-year d1) (date-year d2)) nil)
        ((< (dayofyear d1) (dayofyear d2)) t)
        ((> (dayofyear d1) (dayofyear d2)) nil)
        ((< (minuteofday d1) (minuteofday d2)) t)
        (t nil)))

(defmethod date-equalp ((d1 datetime) (d2 datetime))
  (and (= (date-year d1) (date-year d2))
       (= (date-month d1) (date-month d2))
       (= (date-day d1) (date-day d2))
       (= (date-hour d1) (date-hour d2))
       (= (date-minutes d1) (date-minutes d2))))

(defmethod date-copy ((d datetime))
  (make-date (date-year d) (date-month d) (date-day d) 
             (date-hour d) (date-minutes d)))

(defmethod date-add ((d datetime) minutes)
  (let ((new (date-copy d)))
    (setf (date-minutes new) (+ minutes (date-minutes new)))
    (when (> (date-minutes new) 59)
      ;; Roll over into the next hour.
      (multiple-value-bind (h m) (truncate (date-minutes new) 60)
        (setf (date-hour new) (+ (date-hour new) h))
        (setf (date-minutes new) m))
      (when (> (date-hour new) 23)
        ;; Roll over into the next day.
        (multiple-value-bind (d h) (truncate (date-hour new) 24)
          (setf (date-day new) (+ (date-day new) d))
          (setf (date-hour new) h))
        ;; To do: check that we haven't rolled into a new month.
        ))
    new))
             